<?php

/**
 * JCH Optimize - Performs several front-end optimizations for fast downloads
 *
 * @package   jchoptimize/core
 * @author    Samuel Marshall <samuel@jch-optimize.net>
 * @copyright Copyright (c) 2022 Samuel Marshall / JCH Optimize
 * @license   GNU/GPLv3, or later. See LICENSE file
 *
 *  If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */

namespace JchOptimize\Core\Css\Callbacks;

use JchOptimize\Core\Container\Container;
use JchOptimize\Core\Container\ContainerAwareTrait;
use JchOptimize\Core\Css\CssComponents;
use JchOptimize\Core\Css\Parser;
use JchOptimize\Core\Exception\InvalidArgumentException;
use JchOptimize\Core\Exception\PregErrorException;
use JchOptimize\Core\Exception\PropertyNotFoundException;
use JchOptimize\Core\Registry;
use Joomla\DI\ContainerAwareInterface;

use function defined;

defined('_JCH_EXEC') or die('Restricted access');
abstract class AbstractCallback implements ContainerAwareInterface
{
    use ContainerAwareTrait;

    protected Registry $params;
    protected array $nestingAtRules = ['media', 'supports', 'layer', 'scope', 'container', 'document'];
    private ?Parser $parser = null;
    public function __construct(Container $container, Registry $params)
    {
        $this->container = $container;
        $this->params = $params;
    }
    /**
     * @throws PregErrorException
     */
    public function processMatches(array $matches): string
    {
        if (empty($matches[0])) {
            return $matches[0];
        }
        $nestedRegex = self::cssNestedAtRuleWithCaptureGroupToken();
        if (\preg_match("#^{$nestedRegex}\$#s", $matches[0], $nestedMatches) && \in_array($nestedMatches['name'], $this->nestingAtRules)) {
            $processedContent = $this->getParser()->processMatchesWithCallback($nestedMatches['styleblock'], $this);
            return "@{$nestedMatches['name']} {$nestedMatches['ruleconditions']}{{$processedContent}}";
        }
        $expectedComponents = $this->getExpectedCssComponents();
        foreach ($expectedComponents as $component) {
            try {
                $cssComponent = $component::load($matches[0]);
            } catch (InvalidArgumentException) {
                continue;
            }
            if ($cssComponent instanceof CssComponents) {
                return $this->internalProcessMatches($cssComponent);
            }
        }
        return $matches[0];
    }
    abstract protected function internalProcessMatches(CssComponents $cssComponent): string;
    abstract protected function getExpectedCssComponents(): array;
    public function setParser(Parser $parser): void
    {
        $this->parser = $parser;
    }
    protected function getParser(): Parser
    {
        if (!$this->parser) {
            throw new PropertyNotFoundException('Parser not set');
        }
        return $this->parser;
    }
    private static function cssNestedAtRuleWithCaptureGroupToken(): string
    {
        $esc = Parser::cssEscapedString();
        $dqStr = Parser::doubleQuoteStringToken();
        $sqStr = Parser::singleQuoteStringToken();
        $bc = Parser::blockCommentToken();
        return "@(?:-[^-]++-)?(?<name>[a-zA-Z-]++)\\s*+" . "(?<ruleconditions>(?>[^{}@/\\\\'\";]++|{$bc}|{$esc}|{$dqStr}|{$sqStr}|/)*+){(?<styleblock>.*)}";
    }
}
