<?php

/**
 * JCH Optimize - Performs several front-end optimizations for fast downloads
 *
 * @package   jchoptimize/core
 * @author    Samuel Marshall <samuel@jch-optimize.net>
 * @copyright Copyright (c) 2022 Samuel Marshall / JCH Optimize
 * @license   GNU/GPLv3, or later. See LICENSE file
 *
 *  If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */

namespace JchOptimize\Core\Css\Callbacks;

use CodeAlfa\RegexTokenizer\Debug\Debug;
use DOMNodeList;
use DOMXPath;
use JchOptimize\Core\Css\Components\CssRule;
use JchOptimize\Core\Css\Components\FontFaceAtRule;
use JchOptimize\Core\Css\Components\ImportAtRule;
use JchOptimize\Core\Css\Components\KeyFrames;
use JchOptimize\Core\Css\CssComponents;
use JchOptimize\Core\Exception\PropertyNotFoundException;
use JchOptimize\Core\FeatureHelpers\DynamicSelectors;

use function array_unique;
use function defined;
use function get_class;
use function implode;
use function in_array;
use function preg_last_error;
use function preg_match;
use function preg_replace;
use function preg_replace_callback;
use function preg_split;
use function str_contains;
use function str_replace;
use function str_starts_with;
use function strlen;
use function trim;

use const PREG_NO_ERROR;

defined('_JCH_EXEC') or die('Restricted access');
class ExtractCriticalCss extends \JchOptimize\Core\Css\Callbacks\AbstractCallback
{
    use Debug;

    public string $htmlAboveFold = '';
    public string $fullHtml = '';
    private ?DOMXPath $xPath = null;
    public string $postCss = '';
    public string $preCss = '';
    public bool $isPostProcessing = \false;
    protected string $criticalCss = '';
    protected function internalProcessMatches(CssComponents $cssComponent): string
    {
        $this->_debug($cssComponent->render(), $cssComponent->render(), 'beginExtractCriticalCss');
        if ($cssComponent instanceof FontFaceAtRule || $cssComponent instanceof KeyFrames) {
            if (!$this->isPostProcessing) {
                //If we're not processing font-face or keyframes yet, let's just save them for later until
                // after we've done getting all the // critical css
                $this->postCss .= $cssComponent->render();
            } else {
                if ($cssComponent instanceof FontFaceAtRule) {
                    if (str_contains($this->criticalCss, $cssComponent->getFontFamily())) {
                        //Only include fonts in the critical CSS that are being used above the fold
                        //@TODO prevent duplication of fonts in critical css
                        return $cssComponent->render();
                    }
                } else {
                    // keyframes
                    if (str_contains($this->criticalCss, trim($cssComponent->getName()))) {
                        return $cssComponent->render();
                    }
                }
            }
            return '';
        }
        //We'll prepend imports to the Critical CSS, imported Google font files will never be expanded
        if ($cssComponent instanceof ImportAtRule) {
            $this->preCss .= $cssComponent->render();
        }
        if (!$cssComponent instanceof CssRule) {
            return $cssComponent->render();
        }
        $selectorList = $cssComponent->getSelectorList();
        if (JCH_PRO) {
            /** @see DynamicSelectors::getDynamicSelectors() */
            if ($this->getContainer()->get(DynamicSelectors::class)->getDynamicSelectors($selectorList)) {
                $this->appendToCriticalCss($cssComponent->render());
                $this->_debug('', '', 'afterAddDynamicCss');
                return $cssComponent->render();
            }
            $this->_debug('', '', 'afterSearchDynamicCss');
        }
        $selectors = \explode(',', $selectorList);
        $foundSelectors = array();
        //Iterate through each selector chain
        foreach ($selectors as $selector) {
            //If selector stand-alone pseudo-selector add it
            if (str_starts_with($selector, ':')) {
                $this->appendToCriticalCss($cssComponent->render());
                return $cssComponent->render();
            }
            //If selector is already in critical css just go ahead and add this rule
            if (str_contains($this->criticalCss, $selector)) {
                $this->appendToCriticalCss($cssComponent->render());
                //Retain matched CSS in combined CSS
                return $cssComponent->render();
            }
            //Exclude user-action pseudo-classes
            if (preg_match("#:(?:hover|active|focus(?:\\-visible|\\-within)?)#ix", $selector)) {
                continue;
            }
            //Remove other pseudo selectors except some select input pseudo-classes and functional pseudo-classes
            //$validPseudos = '(?:en|dis)abled|(?:un)?checked|is|not|where|has';
            //Remove pseudo-selectors
            $selector = preg_replace("#::?[a-zA-Z0-9-]++(\\((?>[^()]++|(?1))*\\))?#", '', $selector);
            if (trim($selector) == '') {
                continue;
            }
            //Check CSS selector chain against HTMl above the fold to find a match
            if ($this->checkCssAgainstHtml($selector, $this->htmlAboveFold)) {
                //Match found, add selector to array
                $foundSelectors[] = $selector;
            }
        }
        //If no valid selector was found in the group then we don't add this rule to the critical CSS
        if (empty($foundSelectors)) {
            $this->_debug($selectorList, $cssComponent->render(), 'afterSelectorNotFound');
            //Don't add to critical css
            return '';
        }
        //Group the found selector chains
        $foundSelectorList = implode(',', array_unique($foundSelectors));
        //remove any backslash used for escaping
        //$foundSelectorList = str_replace('\\', '', $foundSelectorList);
        $this->_debug($foundSelectorList, $cssComponent->render(), 'afterSelectorFound');
        $success = null;
        //Convert the selector group to Xpath
        $xPath = $this->convertCss2XPath($foundSelectorList, $success);
        $this->_debug($xPath, $cssComponent->render(), 'afterConvertCss2XPath');
        if ($success !== \false) {
            $element = $this->getDOMXPath()->query($xPath);
            //                                if ($element === FALSE)
            //                                {
            //                                        echo $aMatches[1] . "\n";
            //                                        echo $xPath . "\n";
            //                                        echo $sXPathValue . "\n";
            //                                        echo "\n\n";
            //                                }
            //Match found! Add to critical CSS
            if ($element instanceof DOMNodeList && $element->length) {
                $this->appendToCriticalCss($cssComponent->render());
                $this->_debug($xPath, $cssComponent->render(), 'afterCriticalCssFound');
                return $cssComponent->render();
            }
            $this->_debug($xPath, $cssComponent->render(), 'afterCriticalCssNotFound');
        }
        //No match found for critical CSS.
        return '';
    }
    public function appendToCriticalCss(string $css): void
    {
        $this->criticalCss .= $css;
    }
    /**
     * Do a preliminary simple check to see if a CSS declaration is used by the HTML
     *
     * @param string $selectorChain
     * @param string $html
     *
     * @return   bool   True is all parts of the CSS selector is found in the HTML, false if not
     */
    protected function checkCssAgainstHtml(string $selectorChain, string $html): bool
    {
        //Split selector chain into simple selectors
        $aSimpleSelectors = preg_split('#[^\\[ >+]*+(?:\\[[^\\]]*+\\])?\\K(?:[ >+]*+|$)#', trim($selectorChain), -1, \PREG_SPLIT_NO_EMPTY);
        //We'll do a quick check first if all parts of each simple selector is found in the HTML
        //Iterate through each simple selector
        foreach ($aSimpleSelectors as $sSimpleSelector) {
            //Match the simple selector into its components
            $sSimpleSelectorRegex = '#([_a-z0-9-]*)(?:([.\\#]((?:[_a-z0-9-]|\\\\[^\\r\\n\\f0-9a-z])+))|(\\[((?:[_a-z0-9-]|\\\\[^\\r\\n\\f0-9a-z])+)(?:[~|^$*]?=(?|"([^"\\]]*+)"|\'([^\'\\]]*+)\'|([^\\]]*+)))?\\]))*#i';
            if (preg_match($sSimpleSelectorRegex, $sSimpleSelector, $aS)) {
                //Elements
                if (!empty($aS[1])) {
                    $sNeedle = '<' . $aS[1];
                    //Just include elements that will be generated by the browser
                    $aDynamicElements = array('<tbody');
                    if (in_array($sNeedle, $aDynamicElements)) {
                        continue;
                    }
                    if (!str_contains($html, $sNeedle)) {
                        //Element part of selector not found,
                        //abort and check next selector chain
                        return \false;
                    }
                }
                //Attribute selectors
                if (!empty($aS[4])) {
                    //If the value of the attribute is set we'll look for that
                    //otherwise just look for the attribute
                    $sNeedle = !empty($aS[6]) ? $aS[6] : $aS[5];
                    // . '="';
                    if (!empty($sNeedle) && !str_contains($html, str_replace('\\', '', $sNeedle))) {
                        //Attribute part of selector not found,
                        //abort and check next selector chain
                        return \false;
                    }
                }
                //Ids or Classes
                if (!empty($aS[2])) {
                    $sNeedle = ' ' . $aS[3] . ' ';
                    if (!str_contains($html, str_replace('\\', '', $sNeedle))) {
                        //The id or class part of selector not found,
                        //abort and check next selector chain
                        return \false;
                    }
                }
            }
        }
        //If we get to this point then we've found a simple selector that has all parts in the
        //HTML. Let's save this selector chain and refine its search with Xpath.
        return \true;
    }
    public function convertCss2XPath(string $sSelector, ?bool &$success = null): ?string
    {
        $sSelector = preg_replace('#\\s*([>+~,])\\s*#', '$1', $sSelector);
        $sSelector = trim($sSelector);
        $sSelector = preg_replace('#\\s+#', ' ', $sSelector);
        if ($sSelector === null) {
            $success = \false;
            return null;
        }
        $sSelectorRegex = '#(?!$)' . '([>+~, ]?)' . '([*_a-z0-9-]*)' . '(?:(([.\\#])((?:[_a-z0-9-]|\\\\[^\\r\\n\\f0-9a-z])+))(([.\\#])((?:[_a-z0-9-]|\\\\[^\\r\\n\\f0-9a-z])+))?|' . '(\\[((?:[_a-z0-9-]|\\\\[^\\r\\n\\f0-9a-z])+)(([~|^$*]?=)["\']?([^\\]"\']+)["\']?)?\\]))*' . '#i';
        $result = preg_replace_callback($sSelectorRegex, array($this, 'tokenizer'), $sSelector) . '[1]';
        if (preg_last_error() !== PREG_NO_ERROR) {
            $success = \false;
            return null;
        }
        return $result;
    }
    /**
     * @param string[] $aM
     */
    protected function tokenizer(array $aM): string
    {
        $sXPath = match ($aM[1]) {
            '>' => '/',
            '+' => '/following-sibling::*',
            '~' => '/following-sibling::',
            ',' => '[1] | descendant-or-self::',
            ' ' => '/descendant::',
            default => 'descendant-or-self::',
        };
        if ($aM[1] != '+') {
            $sXPath .= $aM[2] == '' ? '*' : $aM[2];
        }
        if (isset($aM[3]) || isset($aM[9])) {
            $sXPath .= '[';
            $aPredicates = array();
            if (isset($aM[4]) && $aM[4] == '.') {
                $aPredicates[] = "contains(@class, ' " . $aM[5] . " ')";
            }
            if (isset($aM[7]) && $aM[7] == '.') {
                $aPredicates[] = "contains(@class, ' " . $aM[8] . " ')";
            }
            if (isset($aM[4]) && $aM[4] == '#') {
                $aPredicates[] = "@id = ' " . $aM[5] . " '";
            }
            if (isset($aM[7]) && $aM[7] == '#') {
                $aPredicates[] = "@id = ' " . $aM[8] . " '";
            }
            if (isset($aM[9])) {
                if (!isset($aM[11])) {
                    $aPredicates[] = '@' . $aM[10];
                } else {
                    switch ($aM[12]) {
                        case '=':
                            $aPredicates[] = "@{$aM[10]} = ' {$aM[13]} '";
                            break;
                        case '|=':
                            $aPredicates[] = "(@{$aM[10]} = ' {$aM[13]} ' or " . "starts-with(@{$aM[10]}, ' {$aM[13]}'))";
                            break;
                        case '^=':
                            $aPredicates[] = "starts-with(@{$aM[10]}, ' {$aM[13]}')";
                            break;
                        case '$=':
                            $aPredicates[] = "substring(@{$aM[10]}, string-length(@{$aM[10]})-" . strlen($aM[13]) . ") = '{$aM[13]} '";
                            break;
                        case '~=':
                            $aPredicates[] = "contains(@{$aM[10]}, ' {$aM[13]} ')";
                            break;
                        case '*=':
                            $aPredicates[] = "contains(@{$aM[10]}, '{$aM[13]}')";
                            break;
                        default:
                            break;
                    }
                }
            }
            if ($aM[1] == '+') {
                if ($aM[2] != '') {
                    $aPredicates[] = "(name() = '" . $aM[2] . "')";
                }
                $aPredicates[] = '(position() = 1)';
            }
            $sXPath .= implode(' and ', $aPredicates);
            $sXPath .= ']';
        }
        return $sXPath;
    }
    public function setDOMXPath(DOMXPath $xPath): void
    {
        $this->xPath = $xPath;
    }
    protected function getDOMXPath(): DOMXPath
    {
        if ($this->xPath instanceof DOMXPath) {
            return $this->xPath;
        }
        throw new PropertyNotFoundException('DOMXPath not found in ' . get_class($this));
    }
    protected function getExpectedCssComponents(): array
    {
        return [CssRule::class, FontFaceAtRule::class, KeyFrames::class, ImportAtRule::class];
    }
}
