<?php

/**
 * JCH Optimize - Performs several front-end optimizations for fast downloads
 *
 * @package   jchoptimize/core
 * @author    Samuel Marshall <samuel@jch-optimize.net>
 * @copyright Copyright (c) 2022 Samuel Marshall / JCH Optimize
 * @license   GNU/GPLv3, or later. See LICENSE file
 *
 *  If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */

namespace JchOptimize\Core\Css;

use CodeAlfa\RegexTokenizer\Debug\Debug;
use DOMDocument;
use DOMXPath;
use JchOptimize\Core\Container\ContainerAwareTrait;
use JchOptimize\Core\Css\Callbacks\CorrectUrls;
use JchOptimize\Core\Css\Callbacks\ExtractCriticalCss;
use JchOptimize\Core\Css\Callbacks\FormatCss;
use JchOptimize\Core\Css\Callbacks\HandleAtRules;
use JchOptimize\Core\Exception;
use JchOptimize\Core\Exception\PregErrorException;
use JchOptimize\Core\FileInfosUtilsTrait;
use JchOptimize\Core\FileUtils;
use JchOptimize\Core\Html\Processor as HtmlProcessor;
use JchOptimize\Core\Registry;
use JchOptimize\Core\SerializableTrait;
use JchOptimize\Platform\Profiler;
use Joomla\DI\ContainerAwareInterface;
use Psr\Log\LoggerAwareInterface;
use Psr\Log\LoggerAwareTrait;
use Serializable;

use function defined;
use function function_exists;
use function implode;
use function libxml_clear_errors;
use function libxml_use_internal_errors;
use function mb_detect_encoding;
use function preg_replace;

defined('_JCH_EXEC') or die('Restricted access');
class Processor implements LoggerAwareInterface, ContainerAwareInterface, Serializable
{
    use ContainerAwareTrait;
    use LoggerAwareTrait;
    use Debug;
    use FileInfosUtilsTrait;
    use SerializableTrait;

    protected string $css = '';
    /**
     * @var Registry
     */
    private Registry $params;
    /**
     * @var string
     */
    private string $debugUrl = '';
    /**
     * @var CorrectUrls
     */
    private CorrectUrls $correctUrls;
    /**
     * @var ExtractCriticalCss
     */
    private ExtractCriticalCss $extractCriticalCss;
    /**
     * @var FormatCss
     */
    private FormatCss $formatCss;
    /**
     * @var HandleAtRules
     */
    private HandleAtRules $handleAtRules;
    private array $cssInfos = [];
    public function __construct(Registry $params, CorrectUrls $correctUrls, ExtractCriticalCss $extractCriticalCss, FormatCss $formatCss, HandleAtRules $handleAtRules)
    {
        $this->params = $params;
        $this->correctUrls = $correctUrls;
        $this->extractCriticalCss = $extractCriticalCss;
        $this->formatCss = $formatCss;
        $this->handleAtRules = $handleAtRules;
    }
    public function setCssInfos(array $cssInfos): void
    {
        $this->cssInfos = $cssInfos;
        $this->correctUrls->setCssInfos($cssInfos);
        $this->handleAtRules->setCssInfos($cssInfos);
        $this->fileUtils = $this->container->get(FileUtils::class);
        $this->debugUrl = $this->prepareFileUrl($cssInfos, 'css');
        //initialize debug
        $this->_debug($this->debugUrl, '', 'CssProcessorConstructor');
    }
    public function getCss(): string
    {
        return $this->css;
    }
    public function setCss(string $css): void
    {
        if (function_exists('mb_convert_encoding')) {
            $sEncoding = mb_detect_encoding($css);
            if ($sEncoding === \false) {
                $sEncoding = \mb_internal_encoding();
            }
            $css = \mb_convert_encoding($css, 'utf-8', $sEncoding);
        }
        $this->css = $css;
    }
    public function formatCss(): void
    {
        $oParser = new \JchOptimize\Core\Css\Parser();
        $prepareExcl = '\\|"(?>[^"{}|]++|"++)*+"\\|';
        $bc = \JchOptimize\Core\Css\Parser::blockCommentToken();
        $cssRuleList = \JchOptimize\Core\Css\Parser::cssRuleListToken();
        $regularAtRule = \JchOptimize\Core\Css\Parser::cssRegularAtRulesToken();
        $nestedAtRule = \JchOptimize\Core\Css\Parser::cssNestedAtRulesToken();
        $criteria = "(?!\\s++|{$bc}|{$cssRuleList}|{$regularAtRule}|{$nestedAtRule}|{$prepareExcl})";
        $oSearchObject = new \JchOptimize\Core\Css\CssSearchObject();
        $oSearchObject->setCssMatchCriteria($criteria);
        $oSearchObject->setCssMatch(\JchOptimize\Core\Css\Parser::cssInvalidCssToken());
        $oParser->setCssSearchObject($oSearchObject);
        try {
            $this->css = $oParser->processMatchesWithCallback($this->css, $this->formatCss);
        } catch (Exception\PregErrorException $oException) {
            $this->logger->error('FormatCss failed - ' . $this->debugUrl . ': ' . $oException->getMessage());
        }
        $this->_debug($this->debugUrl, '', 'formatCss');
    }
    /**
     * Preload resources in CSS
     *
     * @param string $css Css to process
     * @return void
     */
    public function preloadHttp2(string $css): void
    {
        $this->css = $css;
        $this->processUrls(\true);
    }
    /**
     * The path to the combined CSS files differs from the original path so relative paths to images in the files are
     * converted to absolute paths. This method is used again to preload assets found in the Critical CSS after Optimize
     * CSS Delivery is performed
     *
     * @param bool $isHttp2 Indicates if we're doing the run to preload assets
     * @return void
     */
    public function processUrls(bool $isHttp2 = \false): void
    {
        $oParser = new \JchOptimize\Core\Css\Parser();
        $cssRule = new \JchOptimize\Core\Css\CssSearchObject();
        $cssRule->setCssRuleCriteria(['u' => \JchOptimize\Core\Css\Parser::cssUrlToken()]);
        $cssRule->setCssMatch(\JchOptimize\Core\Css\Parser::cssRuleToken());
        $oParser->setCssSearchObject($cssRule);
        $url = \JchOptimize\Core\Css\Parser::cssUrlToken();
        $dqStr = \JchOptimize\Core\Css\Parser::doubleQuoteStringToken();
        $sqStr = \JchOptimize\Core\Css\Parser::singleQuoteStringToken();
        $atRule = new \JchOptimize\Core\Css\CssSearchObject();
        $atRule->setCssAtRuleCriteria("(?<=@import\\s)");
        $atRule->setCssAtRuleCriteria(['u\'"' => "(?:{$url}|{$dqStr}|{$sqStr})"]);
        $atRule->setCssMatch(\JchOptimize\Core\Css\Parser::cssRegularAtRulesToken());
        $oParser->setCssSearchObject($atRule);
        $nestedAtRule = new \JchOptimize\Core\Css\CssSearchObject();
        $nestedAtRule->setCssMatch(\JchOptimize\Core\Css\Parser::cssNestedAtRulesToken());
        $oParser->setCssSearchObject($nestedAtRule);
        $this->correctUrls->isHttp2 = $isHttp2;
        try {
            $this->css = $oParser->processMatchesWithCallback($this->css, $this->correctUrls);
            $this->css .= $this->correctUrls->getPostCss();
        } catch (Exception\PregErrorException $oException) {
            $sPreMessage = $isHttp2 ? 'Http/2 preload failed' : 'ProcessUrls failed';
            $this->logger->error($sPreMessage . ' - ' . $this->debugUrl . ': ' . $oException->getMessage());
        }
        $this->_debug($this->debugUrl, '', 'processUrls');
    }
    public function processAtRules(): void
    {
        $oParser = new \JchOptimize\Core\Css\Parser();
        $oParser->matchOnlyAtRules();
        $regularAtRule = new \JchOptimize\Core\Css\CssSearchObject();
        $regularAtRule->setCssMatchCriteria('(?=@(?:import|charset))');
        $regularAtRule->setCssMatch(\JchOptimize\Core\Css\Parser::cssRegularAtRulesToken());
        $oParser->setCssSearchObject($regularAtRule);
        $nestedAtRule = new \JchOptimize\Core\Css\CssSearchObject();
        $nestedAtRule->setCssMatchCriteria('(?=@font\\-face)');
        $nestedAtRule->setCssMatch(\JchOptimize\Core\Css\Parser::cssNestedAtRulesToken());
        $oParser->setCssSearchObject($nestedAtRule);
        try {
            $this->css = $this->cleanEmptyMedias($oParser->processMatchesWithCallback($this->css, $this->handleAtRules));
        } catch (Exception\PregErrorException $oException) {
            $this->logger->error('ProcessAtRules failed - ' . $this->debugUrl . ': ' . $oException->getMessage());
        }
        $this->_debug($this->debugUrl, '', 'ProcessAtRules');
    }
    /**
     * @throws PregErrorException
     */
    public function cleanEmptyMedias($css): ?string
    {
        $bc = \JchOptimize\Core\Css\Parser::blockCommentToken();
        $criteria = "(?=@(?>[^{};/]++|{$bc})++{\\s*+})";
        $parser = new \JchOptimize\Core\Css\Parser();
        $parser->matchOnlyAtRules();
        $oCssEmptyMediaObject = new \JchOptimize\Core\Css\CssSearchObject();
        $oCssEmptyMediaObject->setCssMatchCriteria($criteria);
        $oCssEmptyMediaObject->setCssMatch(\JchOptimize\Core\Css\Parser::cssNestedAtRulesToken());
        $parser->setCssSearchObject($oCssEmptyMediaObject);
        return $parser->replaceMatches($css, '');
    }
    public function processConditionalAtRules(): void
    {
        if (!empty($this->cssInfos['media'])) {
            $media = $this->cssInfos['media'];
            $css = $this->css;
            $this->css = "@media {$media}{{$css}}";
        }
        if (!empty($this->cssInfos['supports'])) {
            $supports = $this->cssInfos['supports'];
            $css = $this->css;
            $this->css = "@supports {$supports}{{$css}}";
        }
        if (isset($this->cssInfos['layer'])) {
            $layer = $this->cssInfos['layer'];
            $css = $this->css;
            $this->css = "@layer {$layer}{{$css}}";
        }
        $this->_debug($this->debugUrl, '', 'handleMediaQueries');
    }
    /**
     * @throws Exception\PregErrorException
     */
    public function optimizeCssDelivery($css, $html): string
    {
        !JCH_DEBUG ?: Profiler::start('OptimizeCssDelivery');
        $this->_debug('', '', 'StartCssDelivery');
        //We can't use the $html coming in as argument as that was used to generate cache key. Let's get the
        //HTML from the HTML processor
        /** @var HtmlProcessor $htmlProcessor */
        $htmlProcessor = $this->getContainer()->get(HtmlProcessor::class);
        $html = $htmlProcessor->cleanHtml();
        //Place space around HTML attributes for easy processing with XPath
        $html = preg_replace('#\\s*=\\s*(?|"([^"]*+)"|\'([^\']*+)\'|([^\\s/>]*+))#i', '=" $1 "', $html);
        //Truncate HTML to number of elements set in params
        $sHtmlAboveFold = $htmlProcessor->getAboveFoldHtml($html);
        $this->_debug('', '', 'afterHtmlTruncated');
        $oDom = new DOMDocument();
        //Load HTML in DOM
        libxml_use_internal_errors(\true);
        $oDom->loadHtml($sHtmlAboveFold);
        libxml_clear_errors();
        $oXPath = new DOMXPath($oDom);
        $this->_debug('', '', 'afterLoadHtmlDom');
        $sFullHtml = $html;
        $oParser = new \JchOptimize\Core\Css\Parser();
        $oCssSearchObject = new \JchOptimize\Core\Css\CssSearchObject();
        $oCssSearchObject->setCssMatch(\JchOptimize\Core\Css\Parser::cssRuleToken());
        $oParser->setCssSearchObject($oCssSearchObject);
        $atRuleSearchObject = new \JchOptimize\Core\Css\CssSearchObject();
        $atRuleSearchObject->setCssMatch(\JchOptimize\Core\Css\Parser::cssNestedAtRulesToken());
        $oParser->setCssSearchObject($atRuleSearchObject);
        $atImportSearchObject = new \JchOptimize\Core\Css\CssSearchObject();
        $atImportSearchObject->setCssRuleCriteria('(?=@import)');
        $atImportSearchObject->setCssMatch(\JchOptimize\Core\Css\Parser::cssRegularAtRulesToken());
        $oParser->setCssSearchObject($atImportSearchObject);
        $this->extractCriticalCss->htmlAboveFold = $sHtmlAboveFold;
        $this->extractCriticalCss->fullHtml = $sFullHtml;
        $this->extractCriticalCss->setDOMXPath($oXPath);
        $sCriticalCss = $this->cleanEmptyMedias($oParser->processMatchesWithCallback($css, $this->extractCriticalCss));
        //Process Font-Face and Key frames
        $this->extractCriticalCss->isPostProcessing = \true;
        $preCss = $this->extractCriticalCss->preCss;
        $sPostCss = $oParser->processMatchesWithCallback($this->extractCriticalCss->postCss, $this->extractCriticalCss);
        !JCH_DEBUG ?: Profiler::stop('OptimizeCssDelivery', \true);
        return $preCss . $sCriticalCss . $sPostCss;
        //$this->_debug(self::cssRulesRegex(), '', 'afterCleanCriticalCss');
    }
    public function getImports(): string
    {
        return implode($this->handleAtRules->getImports());
    }
    public function getImages(): array
    {
        return $this->correctUrls->getImages();
    }
    public function getFontFace(): array
    {
        return $this->handleAtRules->getFontFace();
    }
    public function getGFonts(): array
    {
        return $this->handleAtRules->getGFonts();
    }
    public function getPreconnects(): array
    {
        return $this->correctUrls->getPreconnects();
    }
    public function getCssBgImagesSelectors(): array
    {
        return $this->correctUrls->getCssBgImagesSelectors();
    }
    public function getLcpImages(): array
    {
        return $this->correctUrls->getLcpImages();
    }
}
