<?php

/**
 * JCH Optimize - Performs several front-end optimizations for fast downloads
 *
 * @package   jchoptimize/core
 * @author    Samuel Marshall <samuel@jch-optimize.net>
 * @copyright Copyright (c) 2024 Samuel Marshall / JCH Optimize
 * @license   GNU/GPLv3, or later. See LICENSE file
 *
 *  If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */

namespace JchOptimize\Core\Html;

use JchOptimize\Core\Exception\PregErrorException;
use JchOptimize\Core\Html\Elements\BaseElement;

use function preg_match;
use function preg_match_all;
use function preg_replace_callback;

use const PREG_SET_ORDER;

class BuildHtmlElement
{
    protected string $regex = '';
    protected BaseElement $element;
    /**
     * @throws PregErrorException
     */
    public function build(string $html): void
    {
        $elementRegex = self::htmlElementWithCaptureValueToken();
        preg_match("#^{$elementRegex}\$#s", $html, $matches);
        $name = \strtolower($matches['name']);
        $this->element = \JchOptimize\Core\Html\HtmlElementBuilder::$name();
        $attributesRegex = self::htmlAttributeWithCaptureValueToken();
        $attributes = $matches['attributes'] ?? '';
        preg_replace_callback('#' . $attributesRegex . '#ix', [$this, 'loadElementAttributes'], $attributes);
        if (isset($matches['content'])) {
            $this->loadChildren($matches['content']);
        } else {
            $this->element->setOmitClosingTag(\true);
        }
    }
    public function getElement(): \JchOptimize\Core\Html\HtmlElementInterface
    {
        return $this->element;
    }
    public function loadElementAttributes(array $matches): string
    {
        if (empty($matches[0])) {
            return $matches[0];
        }
        $name = $matches['name'];
        $value = $matches['value'] ?? '';
        $delimiter = $matches['delimiter'] ?? '"';
        $this->element->attribute($name, $value, $delimiter);
        return '';
    }
    private static function htmlElementWithCaptureValueToken(): string
    {
        $name = \JchOptimize\Core\Html\Parser::htmlGenericElementToken();
        $attributes = \JchOptimize\Core\Html\Parser::htmlAttributesListToken();
        $endTag = \JchOptimize\Core\Html\Parser::htmlEndTagToken();
        return "<(?<name>{$name})\\b\\s*+(?<attributes>{$attributes}+)/?>(?:(?<content>.*){$endTag})?";
    }
    private static function htmlAttributeWithCaptureValueToken(): string
    {
        return "(?<name>[^\\s/\"'=<>]++)(?:\\s*+=\\s*+(?<delimiter>['\"]?)(?|" . "(?<=[\"])(?<value>(?>[^\"\\\\]++|\\\\.)*+)\"" . "|(?<=['])(?<value>(?>[^'\\\\]++|\\\\.)*+)'" . "|(?<=[=])(?<value>[^\\s>]++)" . "))?";
    }
    /**
     * @throws PregErrorException
     */
    private function loadChildren(string $content): void
    {
        $voidElement = \JchOptimize\Core\Html\Parser::htmlElementToken(null, \true);
        $textElement = \JchOptimize\Core\Html\Parser::htmlElementToken();
        $dqStr = \JchOptimize\Core\Html\Parser::doubleQuoteStringToken();
        $sqStr = \JchOptimize\Core\Html\Parser::singleQuoteStringToken();
        $bc = \JchOptimize\Core\Html\Parser::blockCommentToken();
        $startTag = \JchOptimize\Core\Html\Parser::htmlStartTagToken();
        $voidElementNames = \implode('|', \JchOptimize\Core\Html\HtmlElementBuilder::$voidElements);
        $htmlElementRegex = "(?(?=<(?:{$voidElementNames})){$voidElement}|{$textElement})";
        $regex = "(?:(?>[^<'\"/]++|{$bc}|{$dqStr}|{$sqStr}|/|(?!{$startTag})<)++)" . "|(?:{$htmlElementRegex})";
        preg_match_all("#{$regex}#ix", $content, $matches, PREG_SET_ORDER);
        foreach ($matches as $match) {
            if (preg_match("#^{$htmlElementRegex}\$#ix", $match[0])) {
                $child = \JchOptimize\Core\Html\HtmlElementBuilder::load($match[0]);
                $child->setParent($this->element->getElementName());
                $this->element->addChild($child);
            } else {
                $this->element->addChild($match[0]);
            }
        }
    }
}
