<?php

/**
 * JCH Optimize - Performs several front-end optimizations for fast downloads
 *
 * @package   jchoptimize/core
 * @author    Samuel Marshall <samuel@jch-optimize.net>
 * @copyright Copyright (c) 2022 Samuel Marshall / JCH Optimize
 * @license   GNU/GPLv3, or later. See LICENSE file
 *
 *  If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */

namespace JchOptimize\Core\Css\Callbacks;

use JchOptimize\Core\Combiner;
use JchOptimize\Core\Css\Components\CharsetAtRule;
use JchOptimize\Core\Css\Components\FontFaceAtRule;
use JchOptimize\Core\Css\Components\ImportAtRule;
use JchOptimize\Core\Css\CssComponents;
use JchOptimize\Core\Html\FilesManager;

use function array_merge;
use function defined;
use function extension_loaded;

defined('_JCH_EXEC') or die('Restricted access');
class HandleAtRules extends \JchOptimize\Core\Css\Callbacks\AbstractCallback
{
    private array $atImports = [];
    private array $gFonts = [];
    private array $fontFace = [];
    /**
     * @var array
     */
    private array $cssInfos = [];
    protected function internalProcessMatches(CssComponents $cssComponent): string
    {
        if ($cssComponent instanceof CharsetAtRule) {
            return '';
        }
        if ($cssComponent instanceof FontFaceAtRule) {
            if (!$cssComponent->hasDescriptor('font-display')) {
                $cssComponent->setFontDisplay('swap');
            } elseif ($this->params->get('pro_force_swap_policy', '1')) {
                $cssComponent->setFontDisplay('swap');
            }
            if ($this->params->get('pro_optimizeFonts_enable', '0') && empty($this->cssInfos['combining-fontface'])) {
                $this->fontFace[] = ['content' => $cssComponent->render(), 'media' => $this->cssInfos['media']];
                return '';
            }
        }
        //At this point we should be in import context
        if ($cssComponent instanceof ImportAtRule) {
            $uri = $cssComponent->getUri();
            $media = $cssComponent->getMediaQueriesList();
            //If we're importing a Google font file we may need to optimize it
            if ($this->params->get('pro_optimizeFonts_enable', '0') && $uri->getHost() == 'fonts.googleapis.com') {
                //We have to add Gfonts here so this info will be cached
                $this->gFonts[] = ['url' => $uri, 'media' => $media];
                return '';
            }
            //Don't import Google font files even if replaceImports is enabled
            if (!$this->params->get('replaceImports', '0') || $uri->getHost() == 'fonts.googleapis.com') {
                $this->atImports[] = $cssComponent->render();
                return '';
            }
            /** @var FilesManager $oFilesManager */
            $oFilesManager = $this->getContainer()->get(FilesManager::class);
            if ((string) $uri == '' || $uri->getScheme() == 'https' && !extension_loaded('openssl')) {
                return $cssComponent->render();
            }
            if ($oFilesManager->isDuplicated($uri)) {
                return '';
            }
            $aUrlArray = array();
            $aUrlArray[0]['url'] = $uri;
            $aUrlArray[0]['media'] = $media;
            /** @var Combiner $oCombiner */
            $oCombiner = $this->getContainer()->get(Combiner::class);
            try {
                $importContents = $oCombiner->combineFiles($aUrlArray, 'css');
            } catch (\Exception $e) {
                return $cssComponent->render();
            }
            $this->atImports = array_merge($this->atImports, [$importContents['import']]);
            $this->fontFace = array_merge($this->fontFace, $importContents['font-face']);
            $this->gFonts = array_merge($this->gFonts, $importContents['gfonts']);
            return $importContents['content'];
        }
        return $cssComponent->render();
    }
    public function setCssInfos($cssInfos): void
    {
        $this->cssInfos = $cssInfos;
    }
    public function getImports(): array
    {
        return $this->atImports;
    }
    public function getGFonts(): array
    {
        return $this->gFonts;
    }
    public function getFontFace(): array
    {
        return $this->fontFace;
    }
    protected function getExpectedCssComponents(): array
    {
        return [FontFaceAtRule::class, ImportAtRule::class, CharsetAtRule::class];
    }
}
